(**************************************************************)
(*
 *  Ensemble, (Version 0.61)
 *  Copyright 1999 Cornell University
 *  All rights reserved.
 *
 *  See ensemble/doc/license.txt for further information.
 *)
(**************************************************************)
(**************************************************************)
(* LSET.ML *)
(* Author: Mark Hayden, 11/96 *)
(**************************************************************)
open Util
(**************************************************************)
let ltoa = Arrayf.of_list
let atol = Arrayf.to_list
(**************************************************************)

let sorta a = Arrayf.sort (>=) a
let sort a = atol (sorta (ltoa a))

(**************************************************************)

type state = {
  ft : bool ;
  tf : bool ;
  tt : bool
} 

let rec help1 s l1 l2 =
  match l1,l2 with
  | [],l2 -> if s.ft then l2 else []
  | l1,[] -> if s.tf then l1 else []
  | (h1::t1),(h2::t2) ->
      if h1 < h2 then (
	if s.tf then
	  h1 :: (help1 s t1 l2)
	else
	  (help1 s t1 l2)
      ) else if h1 > h2 then (
	if s.ft then
	  h2 :: (help1 s l1 t2)
	else
	  (help1 s l1 t2)
      ) else (
	if s.tt then
	  h1 :: (help1 s t1 t2)
	else
	  (help1 s t1 t2)
      )

let help2 s l1 l2 =
  let l1 = sort l1 in
  let l2 = sort l2 in
  help1 s l1 l2

let union     = {ft=true ;tf=true ;tt=true}
let intersect = {ft=false;tf=false;tt=true}
let subtract  = {ft=false;tf=true ;tt=false}

let union     l1 l2 = help2 union     l1 l2
let intersect l1 l2 = help2 intersect l1 l2
let subtract  l1 l2 = help2 subtract  l1 l2

(**************************************************************)
(**************************************************************)

let collapse l =
  let l = sort l in
  match l with
  | [] -> []
  | hd::tl ->
      let rec loop last = function
    	| [] -> [last]
    	| hd::tl -> 
	    if last = hd then loop last tl
	    else last :: loop hd tl
      in loop hd tl

(**************************************************************)
(**************************************************************)
(* Debugging versions of above functions.
 *)
(*
let subtract l1 l2 =
  let old_subtract f = function
    | [] -> f
    | e  ->
      	let rec loop = function
          | [] -> []
       	  | elem::l -> if List.mem elem e then loop l else elem :: loop l
      	in loop f
  in
  let ret = subtract l1 l2 in
  if ret <> (sort (old_subtract l1 l2)) then
    failwith "subtract:check" ;
  ret
  
let union l1 l2 =      
  let old_union l1 l2 =
    let rec loop = function
      | [] -> l2
      | a::l -> if List.mem a l2 then loop l else a :: loop l
    in loop l1
  in
  let ret = union l1 l2 in
  if ret <> (sort (old_union l1 l2)) then
    failwith "union:check" ;
  ret

let intersect l1 l2 =  
  let old_intersect l1 l2 =
    let rec loop = function
      | [] -> []
      | a::l -> if List.mem a l2 then a :: loop l else loop l
    in loop l1
  in
  let ret = intersect l1 l2 in
  if ret <> (sort (old_intersect l1 l2)) then
    failwith "union:check" ;
  ret
*)
(**************************************************************)
(**************************************************************)
type 'a t = 'a Arrayf.t
let inject = sorta
let project = Util.ident

let subtracta a b = ltoa (subtract (atol a)(atol b))
let uniona a b = ltoa (union (atol a)(atol b))
let intersecta a b = ltoa (intersect (atol a)(atol b))
let supera a b = subtracta b a = Arrayf.empty
let disjointa a b = intersecta a b = Arrayf.empty
let collapsea a = 
  let a = sorta a in
  let repeat = ref false in
  for i = 1 to pred (Arrayf.length a) do
    if Arrayf.get a (pred i) = Arrayf.get a i then
      repeat := true
  done ;
  if not !repeat then a else (
    ltoa (collapse (atol a))
  )
(**************************************************************)
(*
let subtract a b = atol (subtracta (ltoa a)(ltoa b))
let union a b = atol (uniona (ltoa a)(ltoa b))
let intersect a b = atol (intersecta (ltoa a)(ltoa b))
let collapse a = atol (collapsea (ltoa a))
*)
let super a b = supera (ltoa a) (ltoa b)
let disjoint a b = disjointa (ltoa a) (ltoa b)
(**************************************************************)
